/*
 * Created on 2004/12/18
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package org.hichart.h3code.tool;

import java.util.Iterator;

import org.hichart.h3code.AbstractCell;
import org.hichart.h3code.AbstractLinkedNode;
import org.hichart.h3code.Cell;
import org.hichart.h3code.PerimeterCell;
import org.hichart.h3code.Table;

/**
 * H3-Code Generator 
 * @author Tomokazu ARITA (arita@obirin.ac.jp)
 * @version 0.1.0 2004/12/18 1:07:16 
 *
 */
public class H3CCodeGenerator {
	
	/**
	 * <code>H3Code_CURRENT_VERSION</code> is the H3-Code version. 
	 */
	private static String H3Code_CURRENT_VERSION = "2.0"; //Bug FIX
	
	
	/**
	 * <code>table</code> is the current table. 
	 */
	private Table table = null; //^[Qbg̃e[u
	
	
	/**
	 * <code>pw</code> is a PrintWriter.class instance.
	 */
	private java.io.PrintWriter pw = null;
	
	
	/**
	 * <code>separetor</code> is the separetor of H3-Code  
	 */
	private String separetor = " ";
	
	
	/**
	 * Creates an instance of H3CCodeGenerator.class
	 * @param table  a table
	 * @param pw     a PrintWriter.class instance 
	 */
	public H3CCodeGenerator(Table table, java.io.PrintWriter pw){
		this.table = table;
		this.pw = pw;
	}
	
	/**
	 * Generates a H3-Code file for the setted table. 
	 * @return true (always)
	 */
	public boolean generate(){
		if(table == null){ return false; }
		
		/* e[u̕W */
		normalize();
		
		/* wb_̏o */
		generateHeader(); 
		
		/* Z̏o */
		generateCells();
		
		
		return true;
	}
	
	
	/**
	 * Generates lines for the Cell part. 
	 */
	private void generateCells(){
		java.util.ArrayList cells  = table.getCellSet();
		Iterator it = cells.iterator();
		while(it.hasNext()){
			Object obj = it.next();
			if(obj!=null){ 
				if(obj instanceof AbstractCell){ generateCell((AbstractCell)obj); }
			} //r̗vfnull̏ꍇl
		}
	}
	
	/**
	 * Generates a line for a cell in the current table.
	 * @param cell
	 */
	private void generateCell(AbstractCell cell){
		
		/* (1) output node id */
		pw.print(cell.getID()); pw.print(separetor);
		
		/* (2) output cell type */
		if(cell instanceof PerimeterCell){ pw.print(AbstractCell.PERIMETER_CELL); pw.print(separetor); }
		else if(cell instanceof Cell){ pw.print(AbstractCell.CELL); pw.print(separetor); }
		
		/* (3) output nwe_left */
		pw.print(cell.getConnectedCell(AbstractLinkedNode.NWE_LEFT)); pw.print(separetor);
		
		
		/* (4) output swe_left */
		pw.print(cell.getConnectedCell(AbstractLinkedNode.SWE_LEFT)); pw.print(separetor);
		
		
		/* (5) output nwe_right */
		pw.print(cell.getConnectedCell(AbstractLinkedNode.NWE_RIGHT)); pw.print(separetor);
		
		
		/* (6) output swe_right */
		pw.print(cell.getConnectedCell(AbstractLinkedNode.SWE_RIGHT)); pw.print(separetor);
		
		
		/* (7) output wwe_upper */
		pw.print(cell.getConnectedCell(AbstractLinkedNode.WWE_UPPER)); pw.print(separetor);
		
		
		/* (8) output ewe_upper */
		pw.print(cell.getConnectedCell(AbstractLinkedNode.EWE_UPPER)); pw.print(separetor);
		
		
		/* (9) output wwe_lower */
		pw.print(cell.getConnectedCell(AbstractLinkedNode.WWE_LOWER)); pw.print(separetor);
		
		
		/* (10) output ewe_lower */
		pw.print(cell.getConnectedCell(AbstractLinkedNode.EWE_LOWER)); pw.print(separetor);
		
		
		/* (11) output north wall */
		pw.print(cell.getWall(AbstractCell.NORTH)); pw.print(separetor);
		
		
		/* (12) output south wall */
		pw.print(cell.getWall(AbstractCell.SOUTH)); pw.print(separetor);
		
		
		/* (13) output west wall */
		pw.print(cell.getWall(AbstractCell.WEST)); pw.print(separetor);
		
		
		/* (14) output east wall */
		pw.print(cell.getWall(AbstractCell.EAST)); pw.print(separetor);
		
		
		/* (15) output content id */
		/* 0̂ݑΉ */
		pw.print("0"); pw.print(separetor);
		
		
		/* (16) output content align */
		/* 0̂ݑΉ */
		pw.print("0"); pw.print(separetor);
		
		
		/* (17) output e-point */
		/* 0̂ݑΉ */
		pw.print("0"); pw.print(separetor);
		
		
		/* s */
		pw.println();
	}
	
	/**
	 * Generates a HEAD part of H3-Code. 
	 */
	private void generateHeader(){
		
		/*  */
		java.util.ArrayList cells  = table.getCellSet();
		
		
		/* (1) output version */
		pw.print(H3Code_CURRENT_VERSION); pw.print(separetor);
		
		/* (2) output columns */
		int count = 0;
		AbstractCell cell = table.getHead();
		for(; (cell = (AbstractCell)cells.get(cell.getConnectedCell(AbstractLinkedNode.WWE_LOWER))) instanceof Cell; count++){}
		pw.print(count); pw.print(separetor);
		
		
		/* (2) output rows */
		count = 0;
		cell = table.getHead();
		for(; (cell = (AbstractCell)cells.get(cell.getConnectedCell(AbstractLinkedNode.NWE_LEFT))) instanceof Cell; count++){}
		pw.print(count); pw.print(separetor);
		
		/* s */
		pw.println();
		
	}
	
	/* 폜ɂĐTableXgnullZAIDĐݒ肷 */
	/**
	 * Normalizes the list of cells. 
	 */
	public void normalize(){
		if(table == null){ return; }
		
		java.util.ArrayList cells  = table.getCellSet();
		System.out.println("Table_Size:"+cells.size());//debug
		
		
		/* ID̍Đݒ */
		for(int i=1, count=1; i<cells.size(); i++){
			AbstractCell cell = (AbstractCell)cells.get(i);
			if(cell != null){ cell.setID(count); count++;}
		}
		
		
		/* N̍Ĕzu */
		for(int i=1; i<cells.size(); i++){
			AbstractCell cell = (AbstractCell)cells.get(i);
			if(cell != null){
				
				/* (1) output nwe_left */
				reconnect(cell, AbstractLinkedNode.NWE_LEFT, cells);
				
				
				/* (2) output swe_left */
				reconnect(cell, AbstractLinkedNode.SWE_LEFT, cells);
								
				
				/* (3) output nwe_right */
				reconnect(cell, AbstractLinkedNode.NWE_RIGHT, cells);
				
				
				/* (4) output swe_right */
				reconnect(cell, AbstractLinkedNode.SWE_RIGHT, cells);
				
				
				/* (5) output wwe_upper */
				reconnect(cell, AbstractLinkedNode.WWE_UPPER, cells);
				
				
				/* (6) output ewe_upper */
				reconnect(cell, AbstractLinkedNode.EWE_UPPER, cells);
				
				
				/* (7) output wwe_lower */
				reconnect(cell, AbstractLinkedNode.WWE_LOWER, cells);
				
				
				/* (8) output ewe_lower */
				reconnect(cell, AbstractLinkedNode.EWE_LOWER, cells);
			}
		}
		
		
		/* null̏ */
		for(int i=1; i<cells.size(); i++){
			AbstractCell cell = (AbstractCell)cells.get(i);
			if(cell==null){ cells.remove(i); }
		}
		
		
	}
	
	/**
	 * Reconnect pointers of a cell. 
	 * @param cell      a target cell.
	 * @param position  a pointer position
	 * @param cells     the set of cells in the current table. 
	 */
	private void reconnect(AbstractCell cell, int position, java.util.ArrayList cells){
		AbstractCell cell2; 
		int id;
		
		id = cell.getConnectedCell(position);
		if(id!=0){ 
			cell2 = (AbstractCell)cells.get(id);
			
			if(cell2.getID()!=id){
				cell.disconnect(position); //ڑO͕K
				cell.connect(cell2.getID(), position);
			}
		}
	}
	

}
